(**
   Mersenne Twister for OCaml

   Mersenne Twister
   (http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/MT2002/mt19937ar.html)
   OCamlɖ󂵂̂łB
*)

open Int32

let n = 624
let m = 397
let matrix_a = 0x9908b0dfl
let upper_mask = 0x80000000l
let lower_mask = 0x7fffffffl

(** Ԍ^ *)
type state = {
  mutable counter : int;
  mutable array : int32 array;
}

(** Z̈ɕKvƂȂ obj ĕԂ *)
let create () =
  {
    counter = n;
    array = Array.make n 0l;
  }

(** obj  *)
let init obj seed =
  obj.array.(0) <- seed;
  for n = 1 to (n-1) do
    let pre = obj.array.(n-1) in
    obj.array.(n) <- (add (Int32.of_int n)
                        (mul 1812433253l
                           (logxor pre (shift_right_logical pre 30))));
  done;
  obj.counter <- n
    

(** obj zɂď *)
let init_by_array obj array =
  let get = Array.get obj.array in
  let set = Array.set obj.array in
  
  let alen = (Array.length array) in
  let k = max alen n in
  let i = ref 1 in
  let j = ref 0 in
  
  init obj 19650218l;
  for mk=0 to (k-1) do
    set !i (add (add array.(!j) (Int32.of_int !j))
              (logxor (get !i)
                 (mul 1664525l
                    (logxor (get (!i - 1))
                       (shift_right_logical (get (!i - 1)) 30)))));
    incr i; incr j;
    if (!i >= n) then begin
      set 0 (get (n - 1));
      i := 1
    end;
    if (!j >= alen) then j := 0;
  done;
  for mk=0 to (n-2) do
    set !i (add (neg (Int32.of_int !i))
              (logxor (get !i)
                 (mul 1566083941l
                    (logxor (get (!i - 1))
                       (shift_right_logical (get (!i - 1)) 30)))));
    incr i;
    if (!i >= n) then begin
      set 0 (get (n - 1));
      i := 1
    end;
  done;
  set 0 0x80000000l

let mixbits a b =
  logor (logand a upper_mask) (logand b lower_mask)

let twist a b c =
  ()

(** vZ 32bit int *)
let int32 obj =
  let get = Array.get obj.array in
  let set = Array.set obj.array in
  let logxor3 a b c = (logxor a (logxor b c)) in
  
  let mag01 p = (if ((logand p 1l) > 0l) then matrix_a else 0l) in
  let y = ref 0l in
  if (obj.counter >= n) then begin
    if (obj.counter = (n+1)) then (* default init *)
      init obj 5489l;
    for kk=0 to (n-m-1) do
      y := (logor (logand (get kk) upper_mask)
              (logand (get (kk + 1)) lower_mask));
      set kk (logxor3 (get (kk + m))
                (shift_right_logical !y 1)
                (mag01 !y));
    done;
    for kk=(n-m) to (n-2) do
      y := (logor (logand (get kk) upper_mask)
              (logand (get (kk + 1)) lower_mask));
      set kk (logxor3 (get (kk + m - n))
                (shift_right_logical !y 1)
                (mag01 !y));
    done;
    y := (logor (logand (get (n - 1)) upper_mask)
            (logand (get 0) lower_mask));
    set (n-1) (logxor3 (get (m - 1))
                 (shift_right_logical !y 1)
                 (mag01 !y));
    obj.counter <- 0
  end;
  y := get obj.counter;
  obj.counter <- obj.counter + 1;
  
  y := logxor !y (shift_right_logical !y 11);
  y := logxor !y (logand (shift_left !y 7) 0x9d2c5680l);
  y := logxor !y (logand (shift_left !y 15) 0xefc60000l);
  y := logxor !y (shift_right_logical !y 18);
  !y


(** vZ 31bit ocaml int *)
let int31 obj =
  Int32.to_int (shift_right_logical (int32 obj) 1)

(** vZ 30bit ocaml unsigned int *)
let int30 obj =
  Int32.to_int (shift_right_logical (int32 obj) 2)

(** vZ int \[0,n) *)
let int obj n =
  (int30 obj) mod n

(** vZ real \[0,1\] *)
let real1 obj =
  let v = (Int32.to_float (int32 obj)) *. (1. /. 4294967295.) in
  if v < 0. then v +. 1. else v

(** vZ real \[0,1) *)
let real2 obj =
  let v = (Int32.to_float (int32 obj)) *. (1. /. 4294967296.) in
  if v < 0. then v +. 1. else v

(** vZ real (0,1) *)
let real3 obj =
  let v = (Int32.to_float (int32 obj) +. 0.5) *. (1. /. 4294967296.) in
  if v < 0. then v +. 1. else v

(** vZ real \[0,n) *)
let real obj n =
  (real2 obj) *. n

(** vZ 53bit real \[0,1) *)
let res53 obj =
  let a = Int32.to_float (shift_right_logical (int32 obj) 5) in
  let b = Int32.to_float (shift_right_logical (int32 obj) 6) in
  (a *. 67108864. +. b) *. (1. /. 9007199254740992.)
    



  
    
